function [L,a,b] = rgb2lab(image);
% convert RGB image to CIE-Lab
% one may use the following to read the image from jpg file:
% [image,Map] = imread('image.jpg','jpg');

image = double(image) / double(max(max(max(image))));

M = [0.412453  0.357580  0.189423; ...
     0.212671  0.715160  0.072169; ...
     0.019334  0.119193  0.950227];
white = M*[1 1 1]';

npix = prod(size(image(:,:,1)));
rgb = reshape(image,npix,3)';
XYZ = M*rgb;

ratioX = XYZ(1,:) / white(1);
ratioY = XYZ(2,:) / white(2);
ratioZ = XYZ(3,:) / white(3);

L = zeros(size(XYZ(1,:)));
a = L;
b = L;

ind = find(ratioY > 0.008856);         
val = ratioY(ind) .^ (1/3);
L(ind) =  116 * val - 16 ;     
a(ind) = -500 * val ;
b(ind) =  200 * val ;

ind = find(ratioY <= 0.008856);
val = ratioY(ind);
L(ind) =  903.3 * val;
a(ind) = -500 * (7.787 * val + 16/116);
b(ind) =  200 * (7.787 * val + 16/116);

ind = find(ratioX > 0.008856);
val = ratioX(ind) .^ (1/3);
a(ind) = a(ind) + 500 * val ;
ind = find(ratioX <= 0.008856);
val = ratioX(ind);
a(ind) = a(ind) + 500 * (7.787 * val + 16/116);

ind = find(ratioZ > 0.008856);
val = ratioZ(ind) .^ (1/3);
b(ind) = b(ind) - 200 * val ;
ind = find(ratioZ <= 0.008856);
val = ratioZ(ind);
b(ind) = b(ind) - 200 * (7.787 * val + 16/116);


L = reshape(L,size(image(:,:,1)));
a = reshape(a,size(image(:,:,1)));
b = reshape(b,size(image(:,:,1)));
