#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "s2m.h"

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void histog_init(struct histype *hist, float data)
{
  int it;
  int i,j;
  float av;

  /* get iteration number (=next free bin) */
  it = (int) hist->num[NBINS-1];
  (hist->num[NBINS-1])++;

  /* if first iteration: put data in bin 0 and return */
  if (it==0) {
    hist->top = 0;
    hist->avr[0] = data;
    hist->num[0] = 1;
    hist->order[0] = -1;
    return;
  }

  /* if not first iteration: put data in empty bin, arrange pointers */
  hist->avr[it] = data;
  hist->num[it] = 1;
  av = hist->avr[hist->top];
  if (data < av) {
    hist->order[it] = hist->top;
    hist->top = it;
  } else {    
    j = (int) hist->top;
    i = (int) hist->order[j];
    while (i >= 0) {
      av = hist->avr[i];
      if (data < av)
	break;
      
      j = i;
      i = (int) hist->order[j];
    }
    hist->order[j] = (short)it;
    hist->order[it] = (short)i;
  }
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void histog_insert(struct histype *hist, float data)
{
  int i,j;
  int first, second;
  unsigned short n1,n2;
  float av;

  /* select two bins to be merged */
  minsum(hist, &first, &second);

  /* merge bin (second) into bin (first) */
  n1 = hist->num[first];
  n2 = hist->num[second];
  hist->avr[first] = 
  ((float)n1*hist->avr[first] + (float)n2*hist->avr[second]) / (float)(n1+n2);

  hist->num[first] += hist->num[second];
  hist->order[first] = hist->order[second];

  /* insert new data point into empty bin */
  hist->avr[second] = data;
  hist->num[second] = 1;
  av = hist->avr[hist->top];
  if (data < av) {
    hist->order[second] = hist->top;
    hist->top = second;
  } else {    
    j = (int) hist->top;
    i = (int) hist->order[j];
    while (i >= 0) {
      av = hist->avr[i];
      if (data < av)
	break;
      
      j = i;
      i = (int) hist->order[j];
    }
    hist->order[j] = (short)second;
    hist->order[second] = (short)i;
  }
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void minsum(struct histype *hist, int *first, int *second)
{
  int i, j;
  int minsum, num_sum;

  minsum = INT_MAX;
  j = (int)hist->top;
  i = (int)hist->order[j];
  while (i >= 0) {
    num_sum = (int)(hist->num[i] + hist->num[j]);
    if (num_sum < minsum) {
      minsum = num_sum;
      *first  = j;
      *second = i;
    } else {
      if (num_sum == minsum) {
	if (hist->avr[i]       - hist->avr[j]   < 
	    hist->avr[*second] - hist->avr[*first]) {
	  *first  = j;
	  *second = i;
	}
      }
    }

    j = i;
    i = (int)hist->order[j];
  }    
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
float histog_median(struct histype *hist, int total, int correction)
{
  int i,j;
  int sigma1, sigma2, sigma3;
  float av1, av2, av3, s, m1, m2, m3;
  int t1,t3;
  int sum;

  /* identify the mid bin */
  i = (int)hist->top;
  j = (int)hist->num[i];
  while (i >= 0 && j < (total>>1)) {
    t1 = i; /* needed for the correction below */
    i  = (int)hist->order[i];
    j += (int)hist->num[i];
  }
  /* m1 is the average of the mid bin: */
  m1 = hist->avr[i];
  if (!correction)
    return(m1);

  /* make corrections, if not the first or last bin */
  if (i!=(int)hist->top && (t3=(int)hist->order[i])>=0) {
    sigma1 = j - (int)hist->num[i];
    sigma2 = (int)hist->num[i];
    sigma3 = total-sigma1-sigma2;
    av1 = hist->avr[t1];
    av2 = hist->avr[i] ; /* = m1 */
    av3 = hist->avr[t3];
    s = (float)(sigma3-sigma1) / (float)sigma2;
    m2 = 0.25 * ((av1+2.0*av2+av3) + (av3-av1)*s);
    m3 = av2 + 0.25*(av3-av1)*s;
    if (m3 < 0.0) {
      fprintf(stderr,"warning: m3<0, use m1 instead\n");
      m3 = m1;
    }
    /*
    if (m2 < 0.0) { 
      fprintf(stderr,"warning: m2<0, use m1 instead\n"); 
      m2 = m1; 
    }
    */
  } else {
    m3 = m2 = m1;
  }

  if (correction == 2)
    return(m2);
  else if (correction == 3)
    return(m3);
  else
    error("function is called with wrong correction parameter");
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
struct histype **histog_alloc()
{
  struct histype **jh;
  int i;

  jh = (struct histype **)malloc(NE*sizeof(struct histype *)); 
  if (!jh) error("allocation error");
  for (i=0; i<NE; i++) {
    jh[i] = (struct histype *)malloc(sizeof(struct histype));
    if (!jh[i]) error("allocation error");
    jh[i]->num[NBINS-1] = 0; /* needed by histog_init */
  }

  return(jh);
}
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void histog_print(struct histype *hist)
{
  int i;

  i = hist->top;
  while (i >= 0) {
    printf("%3.1f(%d) ", hist->avr[i], hist->num[i]);
    i = hist->order[i];
  }
  printf("\n");
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
