#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "s2m.h"
/* #include "debug.h" */

/* #define NUMV_STATISTICS_ON */

void init_global(char *fname);
void store_global();
void restore_global();

void graph_contraction(void (*hist_manip)(struct histype *, float), 
		       struct histype **join);
struct xrs *find_crossing(int u1, int u2);
int  select_leaf();
void exclude_leaf(int leaf);

unsigned long **ETree;
struct edge *leaf2ij;
struct vcross **Urep;
struct vcross  *Vrep;
struct xrs *UVleaves;
int *numv;
int *Pname;
int *Psize;
int total_vcross;
int NP; /* number of nodes */
int NE; /* number of edges */
int NL; /* depth of edge tree */

#ifdef NUMV_STATISTICS_ON
  int __i, __max, __min;
  float *__av;
#endif

main(int argc, char *argv[])
{
  char *fname;
  int it, num_it;
  int l;

  struct histype **join;
  int correction;
  float median;
  int imed;




#ifdef DEBUG_MODE
  fprintf(stderr,"\n\n*** debug mode ***\n\n\n");
  if (argc < 3) error("no debug option");
  fname = argv[1];
  init_global(fname);

  switch (atoi(argv[2])) {
  case 1:
    check_select_leaf(atoi(argv[3]));
    break;
  case 2:
    check_exclude_leaf();
    break;
  case 3:
    check_find_crossing();
    break;
  default:
    error("debug option not valid");
    break;
  }
  exit(0);
#endif



  /* command line */
  if (argc != 3) {
    fprintf(stderr,"usage: %s <sparse datafile> <num iterations>\n", argv[0]);
    exit(1);
  }
  fname = argv[1];
  num_it = atoi(argv[2]);
  
  /* initializations */
  init_global(fname);
  store_global();
  join = histog_alloc();

#ifdef NUMV_STATISTICS_ON
    __av = (float *)malloc(NP*sizeof(float));
    if (!__av) error("allocation error");
    __max=0;
    __min=NP;
    for (__i=0; __i<NP; __i++) {
      __av[__i]=0.0;
      fprintf(stderr,"%d ", numv[__i]);
      if (numv[__i] > __max)
	__max = numv[__i];
      if (numv[__i] < __min)
	__min = numv[__i];
      __av[0] += (float)numv[__i];
    }
    fprintf(stderr,"\n");
    fprintf(stderr,"initial numv max = %d\n", __max);
    fprintf(stderr,"initial numv min = %d\n", __min);
    fprintf(stderr,"initial numv avr = %f\n", __av[0] / (float)NP);
    __av[0]=0.0;
    __max=0;
#endif

  /* first NBINS iterations to initialize histograms */
  for (it=0; it<NBINS; it++) {
    fprintf(stderr,"%d ", it);
    graph_contraction(histog_init, join);
    restore_global();
  }
  /* the rest of the desired number of iterations */ 
  for (it=NBINS; it<num_it; it++) {
    if (it%10==0) fprintf(stderr,"%d ", it);
    graph_contraction(histog_insert, join);
    restore_global();
  }

#ifdef NUMV_STATISTICS_ON
  fprintf(stderr,"\nmax numv = %d\n", __max);
  fprintf(stderr,"\naverage numv at each r are printed to stdout...\n");
  for (__i=0; __i<NP; __i++)
    printf("%d %f\n", __i, __av[__i]/(float)num_it);
  exit(0);
#endif

  /* compute medians */
  correction = 0;
  fprintf(stderr, "use correction=%d for median estimation\n", correction);
  for (l=0; l<NE; l++) {
    median = histog_median(join[l], num_it, correction);
    imed = (int)median;
    if (median - (float)imed > 0.5)
      imed++;
    /* printf("%f %d\n", median, imed); */
    printf("%d\n", imed);

    /*
      histog_print(join[l]);
    */
  }

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void graph_contraction(void (*hist_manip)(struct histype *, float), 
		       struct histype **join)
{
  int u,v;
  float r;
  int leaf;
  struct xrs *xedge;

  r = (float)NP;
  while (total_vcross > 0) {
    leaf = select_leaf();
    r--;
    
    u = Pname[leaf2ij[leaf].u];
    while (u != Pname[u])
      u = Pname[u];
    
    v = Pname[leaf2ij[leaf].v];
    while (v != Pname[v])
      v = Pname[v];

    xedge = find_crossing(u,v);

#ifdef NUMV_STATISTICS_ON
    __av[(int)r] += (float)total_vcross / r;
    for (__i=0; __i<NP; __i++) {
      if (numv[__i] > __max) {
	__max = numv[__i];
      }
    }
#endif

    while (xedge != NULL) {
      leaf = xedge->leaf;
      exclude_leaf(leaf);
      (*hist_manip)(join[leaf], r);
      xedge = xedge->next;
    }
  }
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void init_global(char *fname)
{
  FILE *pf;
  int i;
  int  tree_alloc (int ne);
  void read_data(FILE *pf);
  void prop_weight();
  void init_reps2l();

  if ((pf=fopen(fname,"r"))==NULL)
    error("cannot open input file");
  fscanf(pf,"%*s %d %d", &NP, &NE);

  /* memory allocations */
  NL = tree_alloc(NE);
  leaf2ij = (struct edge *)malloc((NE+NE%2)*sizeof(struct edge));
  if (!leaf2ij) error("allocation error");
  Urep = (struct vcross **)malloc(NP*sizeof(struct vcross *));
  if (!Urep) error("allocation error");
  Vrep = (struct vcross *)malloc(NE*sizeof(struct vcross));
  if (!Vrep) error("allocation error");
  UVleaves = (struct xrs *)malloc(NE*sizeof(struct xrs));
  if (!UVleaves) error("allocation error");
  numv = (int *)malloc(NP*sizeof(int));
  if (!numv) error("allocation error");
  Pname = (int *)malloc(NP*sizeof(int));
  if (!Pname) error("allocation error");
  Psize = (int *)malloc(NP*sizeof(int));
  if (!Psize) error("allocation error");


  /* initial values */
  total_vcross = NE;
  for (i=0; i<NP; i++) {
    numv[i]=0;
    Pname[i]=i;
    Psize[i]=1;
  }

  read_data(pf); /* initalize ETree (leaves level) and leaf2ij */
  fclose(pf);

  prop_weight(); /* initialize the rest of ETree */
  
  init_reps2l(); /* initialize Urep, Vrep, UVleaves*/
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void init_reps2l()
{
  int u, v, tmp, n;
  struct vcross *p, **pp;

  for (n=0; n<NP; n++)
    Urep[n] = NULL;

  for (n=0; n<NE; n++) {
    u = leaf2ij[n].u;
    v = leaf2ij[n].v;

    /* try to balance the length of v-lists */
    if (numv[u] > numv[v]) {
      tmp = u;
      u = v;
      v = tmp;
    }

    /* enter the name (v) and the leaf (n) into Vrep and UVleaves */
    p = Vrep+n;
    p->vn = v;
    p->next = NULL;
    p->leaves = p->tail = UVleaves+n;

    p->leaves->leaf = n;
    p->leaves->next = NULL;


    /* put the new v-record at the beginning of the u-list */
    numv[u]++;
    p->next = Urep[u];
    Urep[u] = p;
    
    /* put the new v-record at the end of the u-list */
    /*
    numv[u]++;
    pp = &(Urep[u]);
    while ((*pp) != NULL)
      pp = &((*pp)->next);
    *pp = p;
    */
  }

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* get u1 and u2 to be merged.
   exchange u1 and u2 if Psize[u1] < Psize[u2].
   create a new list under u1 and delete u2, also update Pname and Psize.
   return a pointer to a list of crossing edges between u1 and u2. */
struct xrs *find_crossing(int u1, int u2)
{
  int order_vcross(const void *x, const void *y);
  struct vcross *wksp, *p, *vloc;
  struct xrs *p2cross = NULL;
  int r, i, count;
  int u, vn, name;
  int indx, tmp;

  /* always make the smaller tree a child of the larger in Pname */
  if (Psize[u1] < Psize[u2]) {
    tmp = u1;
    u1 = u2;
    u2 = tmp;
  }
  Psize[u1] += Psize[u2];
  Psize[u2] = 0;

  /* decrease total_vcross (will be increased below) */
  total_vcross -= (numv[u1]+numv[u2]);

  /* create workspace */
  wksp = (struct vcross *)malloc((numv[u1]+numv[u2])*sizeof(struct vcross));
  if (!wksp) error("allocation error");

  /* correct the vn fields of each u-list and copy to wksp  */
  count = 0;
  for (i=0; i<2; i++) {
    if (i==0) u=u1; else u=u2;
    p = Urep[u];
    while (p != NULL) {
      /* find names of representatives */
      vn = Pname[p->vn];
      while (vn != Pname[vn])
	vn = Pname[vn];

      /* update Pname (may help in the future) */
      name = (vn==u2) ? u1 : vn;
      indx = p->vn;
      while (Pname[indx] != vn) {
	tmp = Pname[indx];
	Pname[indx] = name;
	indx = tmp;
      }

      /* copy item into wksp */
      /* a label u1 is changed to u2, in order to have the same name for 
	 crossing edges in both lists */
      name = (vn==u1) ? u2 : vn;
      wksp[count].vn = name;
      wksp[count].leaves = p->leaves;
      wksp[count].tail   = p->tail;
      wksp[count].next   = p; /* use the "next" field to store self address */
      count++;
      p = p->next;
    }
  }

  /* sort wksp, concatenate lists with the same name */
  qsort(wksp, count, sizeof(struct vcross), order_vcross);
  r=0;
  while (r < count) {
    if ((i=r+1)==count)
      break;
    while (wksp[i].vn == wksp[r].vn) {
      p = wksp[r].next; /* the real address in Vrep of item wksp[r] */
      p->tail->next = wksp[i].leaves;
      p->tail = wksp[i].tail;
      if (++i == count)
	break;
    }
    r=i;
  }

  /* build the list of u1, store a pointer to crossing edges */
  /* transfer records to other lists if they are shorter */
  Urep[u1] = NULL;
  numv[u1] = 0;		  
  r=0;
  while (r < count) {
    name = wksp[r].vn;
    vloc = wksp[r].next;

    if (name != u2) { /* the v-record is not the crossing one */

      if (numv[u1] <= numv[name]) {
	/* add the record into u1 list */ 
	vloc->next = Urep[u1];
	Urep[u1]   = vloc;
	
	Urep[u1]->vn = name;
	numv[u1]++;
	total_vcross++;
      } else {
	/* add the record into v-list */
	vloc->next = Urep[name];
	Urep[name] = vloc;

	Urep[name]->vn = u1;
	numv[name]++;
	total_vcross++;
      }

    } else { /* the cross u1-u2 */
      if (p2cross != NULL)
	error("p2cross is already assigned");
      p2cross = wksp[r].leaves;
    }

    if ((i=r+1)==count)
      break;
    while (wksp[i].vn == name) {
      if (++i == count)
	break;
    }
    r=i;
  }

  /* u2 is no longer a valid name */
  numv[u2] = 0;
  Urep[u2] = NULL;
  Pname[u2] = u1;

  free(wksp);
  return(p2cross);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int order_vcross(const void *x, const void *y)
{
  struct vcross *a, *b;

  a = (struct vcross *)x;
  b = (struct vcross *)y;

  /* sort by increasing order */
  if (a->vn < b->vn) return (-1);
  if (a->vn > b->vn) return ( 1);
  return(0);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* get number of leaves (ne), allocate tree, return number of levels (nl). */
/* each tree level contains EVEN NUMBER of nodes */
int tree_alloc (int ne)
{
  int k,l,nl;

  /* find number of levels in the tree */
  nl = 1;
  k = ne;
  while (k>1) {
    nl++;
    if (k%2)
      k++;

    k >>= 1;
  }

  /* allocate nl levels, each has even length */
  ETree = (unsigned long **)malloc(nl*sizeof(unsigned long *));
  if (!ETree) error("allocation error");

  l = nl;
  k = ne;
  while (k>1) {
    l--;
    if (k%2)
      k++;

    ETree[l] = (unsigned long *)malloc(k*sizeof(unsigned long));
    if (!ETree[l]) error("allocation error");

    k >>= 1;
  }
  ETree[0] = (unsigned long *)malloc(sizeof(unsigned long));
  if (!ETree[0]) error("allocation error");

  return(nl);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void read_data(FILE *pf)
{
  int i, j, w;
  int last, n;
  
  last = NL-1;
  for (n=0; n<NE; n++) {
    fscanf(pf,"%d %d %d", &i, &j, &w);
    ETree[last][n] = w;
    leaf2ij[n].u = i;
    leaf2ij[n].v = j;
  }

  if (NE%2) {
    ETree[last][NE] = 0;
    leaf2ij[NE].u = -1; /* not used */
    leaf2ij[NE].v = -1; /* not used */
  }
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void prop_weight()
{
  int i,k,l,p,q;

  k = NE;
  if (k%2)
    k++;
  for (l=NL-2; l>0; l--) {
    k >>= 1;
    for(i=0; i<k; i++) {
      p=i<<1;
      q=p+1;
      if (ETree[l+1][p] > ULONG_MAX>>1 || ETree[l+1][q] > ULONG_MAX>>1)
	error("danger of overflow in ETree initialization");
      ETree[l][i] = ETree[l+1][p] + ETree[l+1][q];
    }
    if (k%2) {
      ETree[l][k] = 0;
      k++;
    }
  }
  ETree[0][0] = ETree[1][0] + ETree[1][1];
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
int select_leaf()
{
  int n,l,p,q;
  unsigned long sum, coin;
  unsigned long x, y;

  n = 0;
  for (l=1; l<NL; l++) {
    p = n<<1;
    q = p+1;

    /* the tree is filled with zeros; worth to check explicitly for zero: */
    if ((x=ETree[l][p]) == 0)
      n=q;
    else if ((y=ETree[l][q]) == 0)
      n=p;
    else {
      sum = x+y;
      /* select coin from {0,1,2, ... ,(sum-1)} */
      do 
	coin = (unsigned long)((double)sum *((double)rand()/(double)RAND_MAX));
      while (coin >= sum);
      
      if (coin < x)
	n=p;
      else
	n=q;
    }
  }

  return(n);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void exclude_leaf(int leaf)
{
  unsigned long weight;
  int l, n;

  weight = ETree[NL-1][leaf];
  n = leaf;
  for (l=NL-1; l>=0; l--) {
    ETree[l][n] -= weight;
    n >>= 1;
  }
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
void error(char *message)
{
  fprintf(stderr,"Quit due to: %s\n", message);
  exit(1);
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

